@@ -1,59 +1,27 @@
-# Primitive tests on cluster-enabled redis with modules using redis-cli
+# Primitive tests on cluster-enabled redis with modules
 
 source tests/support/cli.tcl
 
-set testmodule [file normalize tests/modules/blockonkeys.so]
-set testmodule_nokey [file normalize tests/modules/blockonbackground.so]
-set testmodule_blockedclient [file normalize tests/modules/blockedclient.so]
-
-# make sure the test infra won't use SELECT
-set old_singledb $::singledb
-set ::singledb 1
-
 # cluster creation is complicated with TLS, and the current tests don't really need that coverage
 tags {tls:skip external:skip cluster modules} {
 
-# start three servers
-set base_conf [list cluster-enabled yes cluster-node-timeout 1 loadmodule $testmodule]
-start_server [list overrides $base_conf] {
-start_server [list overrides $base_conf] {
-start_server [list overrides $base_conf] {
+set testmodule_nokey [file normalize tests/modules/blockonbackground.so]
+set testmodule_blockedclient [file normalize tests/modules/blockedclient.so]
+set testmodule [file normalize tests/modules/blockonkeys.so]
+
+set modules [list loadmodule $testmodule loadmodule $testmodule_nokey loadmodule $testmodule_blockedclient]
+start_cluster 3 0 [list config_lines $modules] {
 
     set node1 [srv 0 client]
     set node2 [srv -1 client]
     set node3 [srv -2 client]
     set node3_pid [srv -2 pid]
 
-    # the "overrides" mechanism can only support one "loadmodule" directive
-    $node1 module load $testmodule_nokey
-    $node2 module load $testmodule_nokey
-    $node3 module load $testmodule_nokey
-
-    $node1 module load $testmodule_blockedclient
-    $node2 module load $testmodule_blockedclient
-    $node3 module load $testmodule_blockedclient
-
-    test {Create 3 node cluster} {
-        exec src/redis-cli --cluster-yes --cluster create \
-                           127.0.0.1:[srv 0 port] \
-                           127.0.0.1:[srv -1 port] \
-                           127.0.0.1:[srv -2 port]
-
-        wait_for_condition 1000 50 {
-            [csi 0 cluster_state] eq {ok} &&
-            [csi -1 cluster_state] eq {ok} &&
-            [csi -2 cluster_state] eq {ok}
-        } else {
-            fail "Cluster doesn't stabilize"
-        }
-    }
-
     test "Run blocking command (blocked on key) on cluster node3" {
         # key9184688 is mapped to slot 10923 (first slot of node 3)
         set node3_rd [redis_deferring_client -2]
         $node3_rd fsl.bpop key9184688 0
         $node3_rd flush
-
         wait_for_condition 50 100 {
             [s -2 blocked_clients] eq {1}
         } else {
@@ -164,11 +132,11 @@ start_server [list overrides $base_conf] {
 
     test "Kill a cluster node and wait for fail state" {
         # kill node3 in cluster
-        exec kill -SIGSTOP $node3_pid
+        pause_process $node3_pid
 
         wait_for_condition 1000 50 {
-            [csi 0 cluster_state] eq {fail} &&
-            [csi -1 cluster_state] eq {fail}
+            [CI 0 cluster_state] eq {fail} &&
+            [CI 1 cluster_state] eq {fail}
         } else {
             fail "Cluster doesn't fail"
         }
@@ -190,15 +158,65 @@ start_server [list overrides $base_conf] {
         assert_error "ERR Can not execute a command 'set' while the cluster is down" {$node1 do_rm_call set x 1}
     }
 
-    exec kill -SIGCONT $node3_pid
+    resume_process $node3_pid
     $node1_rd close
     $node2_rd close
-
-# stop three servers
-}
-}
 }
 
-} ;# tags
+set modules [list loadmodule [file normalize tests/modules/keyspace_events.so]]
+start_cluster 2 2 [list config_lines $modules] {
 
-set ::singledb $old_singledb
+    set master1 [srv 0 client]
+    set master2 [srv -1 client]
+    set replica1 [srv -2 client]
+    set replica2 [srv -3 client]
+    
+    test "Verify keys deletion and notification effects happened on cluster slots change are replicated inside multi exec" {
+        $master2 set count_dels_{4oi} 1
+        $master2 del count_dels_{4oi}
+        assert_equal 1 [$master2 keyspace.get_dels]
+        assert_equal 1 [$replica2 keyspace.get_dels]
+        $master2 set count_dels_{4oi} 1
+        
+        set repl [attach_to_replication_stream_on_connection -3]
+
+        $master1 cluster bumpepoch
+        $master1 cluster setslot 16382 node [$master1 cluster myid]
+
+        wait_for_cluster_propagation
+        wait_for_condition 50 100 {
+            [$master2 keyspace.get_dels] eq 2
+        } else {
+            fail "master did not delete the key"
+        }
+        wait_for_condition 50 100 {
+            [$replica2 keyspace.get_dels] eq 2
+        } else {
+            fail "replica did not increase del counter"
+        }
+
+        assert_replication_stream $repl {
+            {multi}
+            {del count_dels_{4oi}}
+            {keyspace.incr_dels}
+            {exec}
+        }
+        close_replication_stream $repl
+    }
+}
+
+}
+
+set testmodule [file normalize tests/modules/basics.so]
+set modules [list loadmodule $testmodule]
+start_cluster 3 0 [list config_lines $modules] {
+    set node1 [srv 0 client]
+    set node2 [srv -1 client]
+    set node3 [srv -2 client]
+
+    test "Verify RM_Call inside module load function on cluster mode" {
+        assert_equal {PONG} [$node1 PING]
+        assert_equal {PONG} [$node2 PING]
+        assert_equal {PONG} [$node3 PING]
+    }
+}
