@@ -43,6 +43,9 @@
 #
 # /!\ This script should be run as root
 #
+# NOTE: This script will not work on Mac OSX.
+#       It supports Debian and Ubuntu Linux.
+#
 ################################################################################
 
 die () {
@@ -70,6 +73,16 @@ if [ "$(id -u)" -ne 0 ] ; then
 	exit 1
 fi
 
+#bail if this system is managed by systemd
+_pid_1_exe="$(readlink -f /proc/1/exe)"
+if [ "${_pid_1_exe##*/}" = systemd ]
+then
+	echo "This systems seems to use systemd."
+	echo "Please take a look at the provided example service unit files in this directory, and adapt and install them. Sorry!"
+	exit 1
+fi
+unset _pid_1_exe
+
 if ! echo $REDIS_PORT | egrep -q '^[0-9]+$' ; then
 	_MANUAL_EXECUTION=true
 	#Read the redis port
